TEXT_VERBOSE = False
PYLAB_INTERACTIVE = True
PYLAB_VERBOSE = False
BASE_FIG_NUM = 60
LOG_DATA = True
LOG_DIR = '${DEFAULT}/z_piezo_utils'

NULL_STEPS_BEFORE_SINGLE_PT_SWEEP = False
SLEEP_DURING_SURF_POS_AQUISITION = False # doesn't help

from numpy import array, arange, ones, zeros, linspace, uint16, float, sin, cos, pi
from scipy.stats import linregress
import curses_check_for_keypress
import scipy.optimize
import data_logger
import time

_flush_plot = None
_final_flush_plot = None
_pylab = None
def _dummy_fn(): pass

def _import_pylab():
    """Import pylab plotting functions for when we need to plot.  This
    function can be called multiple times, and ensures that the pylab
    setup is only imported once.  It defines the functions
      _flush_plot() to be called after incremental changes,
      _final_flush_plot(), to be called at the end of any graphical processing, and
      _pylab, a reference to the pylab module."""
    global _pylab
    global _flush_plot
    global _final_flush_plot
    if _pylab == None :
        import pylab as _pylab
    if _flush_plot == None :
        if PYLAB_INTERACTIVE :
            _flush_plot = _pylab.draw
        else :
            _flush_plot = _pylab.show
    if _final_flush_plot == None :
        if PYLAB_INTERACTIVE :
            _final_flush_plot = _pylab.draw
        else :
            _final_flush_plot = _dummy_fn

class error (Exception) :
    pass
class poorFit (error) :
    pass
class tooClose (error) :
    pass

def moveToPosOrDef(zpiezo, pos, defl, step, return_data=False) :
    if return_data or LOG_DATA or PYLAB_VERBOSE :
        aquire_data = True
    else :
        aquire_data = False
    zpiezo._check_range(pos)
    if step == 0 :
        raise error, "Must have non-zero step size"
    elif step < 0 and pos > zpiezo.curPos() :
        step = -step
    elif step > 0 and pos < zpiezo.curPos() :
        step = -step
        
    zpiezo._jump.reserve()
    zpiezo.updateInputs()
    if TEXT_VERBOSE :
        print "current z_piezo position %6d, current deflection %6d <? %6d" % (zpiezo.curPos(), zpiezo.curDef(), defl)
    if aquire_data :
        def_array=[zpiezo.curDef()]
        pos_array=[zpiezo.curPos()]
    if NULL_STEPS_BEFORE_SINGLE_PT_SWEEP == True :
        # take a few null steps to let the input deflection stabalize
        for i in range(100) :
            zpiezo.jumpToPos(zpiezo.curPos())
            if aquire_data :
                def_array.append(zpiezo.curDef())
                pos_array.append(zpiezo.curPos())            
    # step in until we hit our target position or exceed our target deflection
    while zpiezo.curPos() != pos and zpiezo.curDef() < defl :
        distTo = pos - zpiezo.curPos()
        if abs(distTo) < abs(step) :
            jumpTo = pos
        else :
            jumpTo = zpiezo.curPos() + step
        zpiezo.jumpToPos(jumpTo)
        if TEXT_VERBOSE :
            print "current z_piezo position %6d, current deflection %6d >=< %6d" % (zpiezo.curPos(), zpiezo.curDef(), defl)
        if aquire_data :
            def_array.append(zpiezo.curDef())
            pos_array.append(zpiezo.curPos())
    zpiezo._jump.release()
    
    if zpiezo.setExternalZPiezo != None :
        zpiezo.setExternalZPiezo(zpiezo.pos_out2nm(zpiezo.curPos()))
    if PYLAB_VERBOSE :
        _import_pylab()
        _pylab.figure(BASE_FIG_NUM)
        timestamp = time.strftime('%H%M%S')
        _pylab.plot(pos_array, def_array, '.', label=timestamp)
        _pylab.title('step approach')
        _pylab.legend(loc='best')
        _flush_plot()
    if LOG_DATA :
        log = data_logger.data_log(LOG_DIR, noclobber_logsubdir=False,
                                   log_name="moveToPosOrDef")
        data = {"Z piezo output":array(pos_array),
                "Deflection input":array(def_array)}
        log.write_dict_of_arrays(data)
    if return_data :
        data = {"Z piezo output":array(pos_array),
                "Deflection input":array(def_array)}
        return (zpiezo.curVals(), data)
    else :
        return zpiezo.curVals()


def wiggleForInterferenceMin(zpiezo, wig_amp=20000, wig_freq=100,
                             plotVerbose=True) :
    "Output sin to measure interference"
    if PYLAB_VERBOSE or plotVerbose:
        _import_pylab()
    npoints = 1000
    scanfreq = wig_freq*npoints
    out = zeros((npoints,), dtype=uint16)
    for i in range(npoints) :
        out[i] = int(sin(4*pi*i/npoints)*wig_amp+32000)
        # 4 for 2 periods
    if PYLAB_VERBOSE or plotVerbose:
        _pylab.figure(BASE_FIG_NUM+1)
        a = _pylab.axes()
        _pylab.title('wiggle for interference')
        _pylab.hold(False)
        p = _pylab.plot(out, out, 'b.-')
    if LOG_DATA :
        log = data_logger.data_log(LOG_DIR, noclobber_logsubdir=False,
                                   log_name="wiggle")
    c = curses_check_for_keypress.check_for_keypress(test_mode=False)
    while c.input() == None :
        data = zpiezo.ramp(out, scanfreq)
        dmin = data["Deflection input"].min()
        dmax = data["Deflection input"].max()
        if PYLAB_VERBOSE or plotVerbose:
            p[0].set_ydata(data["Deflection input"])
            a.set_ylim([dmin,dmax])
            _flush_plot()
        if LOG_DATA :
            log.write_dict_of_arrays(data)
    if PYLAB_VERBOSE or plotVerbose:
        _pylab.hold(True)

def getSurfPosData(zpiezo, maxDefl, textVerbose=False, plotVerbose=False) :
    "Measure the distance to the surface"
    if plotVerbose and not PYLAB_VERBOSE :
        _import_pylab()
    origPos = zpiezo.curPos()
    # fully retract the piezo
    if TEXT_VERBOSE or textVerbose :
        print "\tRetract the piezo"
    out = linspace(zpiezo.curPos(), zpiezo.zpMin, 2000)
    ret1 = zpiezo.ramp(out, 10e3)
    del(out) # start with a clean output vector, seems to help reduce bounce?
    zpiezo.updateInputs() # opening output seems to cause bounce?
    # locate high deflection position
    if TEXT_VERBOSE or textVerbose :
        print "\tApproach until there is dangerous deflection (> %d)" % maxDefl
    if SLEEP_DURING_SURF_POS_AQUISITION == True :
        time.sleep(.2) # sleeping briefly seems to reduce bounce?
    cp, mtpod = moveToPosOrDef(zpiezo, zpiezo.zpMax, maxDefl,
                               (zpiezo.zpMax-zpiezo.zpMin)/2000,
                               return_data=True)
    highContactPos = zpiezo.curPos()
    if highContactPos <= origPos :   # started out too close, we need some room to move in...
        if TEXT_VERBOSE or textVerbose :
            print "\tToo close, return to the original position: ", origPos
        out = linspace(zpiezo.curPos(), origPos, 2000)
        ret3=zpiezo.ramp(out, 10e3)
        del(out)
        zpiezo.updateInputs()
        if PYLAB_VERBOSE or plotVerbose :
            _pylab.figure(BASE_FIG_NUM+2)
            def plot_dict(d, label) :
                print d.keys()
                _pylab.plot(d["Z piezo output"], d["Deflection input"], '.',label=label)
            plot_dict(ret1, 'First retract')
            plot_dict(mtpod, 'Single step in')
            plot_dict(ret3, 'Return to start')
            _pylab.legend(loc='best')
            _pylab.title('Too close')
        raise tooClose, "Way too close!\n(highContactPos %d <= origPos %d)" % (highContactPos, origPos) 
    zpiezo.updateInputs()
    # fully retract the piezo again
    if TEXT_VERBOSE or textVerbose :
        print "\tRetract the piezo again"
    if SLEEP_DURING_SURF_POS_AQUISITION == True :
        time.sleep(.2)
    out = linspace(highContactPos, zpiezo.zpMin, 2000)
    ret2 = zpiezo.ramp(out, 10e3)
    del(out)
    zpiezo.updateInputs()
    if zpiezo.curPos() != zpiezo.zpMin :
        raise Exception, "bad code"
    # scan to the high contact position
    if TEXT_VERBOSE or textVerbose :
        print "\tRamp in to the deflected position: ", highContactPos
    if SLEEP_DURING_SURF_POS_AQUISITION == True :
        time.sleep(.2)
    out = linspace(zpiezo.curPos(), highContactPos, 4000)
    data = zpiezo.ramp(out, 10e3)
    del(out)
    zpiezo.updateInputs()
    if LOG_DATA :
        log = data_logger.data_log(LOG_DIR, noclobber_logsubdir=False,
                                   log_name="getSurfPos")
        log.write_dict_of_arrays(data)
    if SLEEP_DURING_SURF_POS_AQUISITION == True :
        time.sleep(.2)
    # return to the original position
    if TEXT_VERBOSE or textVerbose :
        print "\tReturn to the original position: ", origPos
    out = linspace(highContactPos, origPos, 2000)
    ret3=zpiezo.ramp(out, 10e3)
    del(out)
    zpiezo.updateInputs()
    return {"ret1":ret1, "mtpod":mtpod, "ret2":ret2,
            "approach":data, "ret3":ret3}

def bilinear(x, params):
    """bilinear fit for surface bumps.  Model has two linear regimes
    which meet at x=kink_position and have independent slopes.
    """
    assert type(x) == type(array([0,1])), "Invalid x array: %s" % type(x)
    left_offset,left_slope,kink_position,right_slope = params
    left_mask = x < kink_position
    right_mask = x >= kink_position # = not left_mask
    left_y = left_offset + left_slope*x
    right_y = left_offset + left_slope*kink_position \
        + right_slope*(x-kink_position)
    return left_mask * left_y + right_mask * right_y

def analyzeSurfPosData(ddict, textVerbose=False, plotVerbose=False, retAllParams=False) :
    # ususes ddict["approach"] for analysis
    # the others are just along to be plotted

    data = ddict["approach"]
    # analyze data, using bilinear model
    # y = p0 + p1 x                for x <= p2
    #   = p0 + p1 p2 + p3 (x-p2)   for x >= p2
    if TEXT_VERBOSE or textVerbose :
        print "\tAnalyze the data"
    dump_before_index = 0 # 25 # HACK!!
    # Generate a reasonable guess...
    startPos = int(data["Z piezo output"].min())
    finalPos = int(data["Z piezo output"].max())
    startDef = int(data["Deflection input"].min())
    finalDef = int(data["Deflection input"].max())
    # startDef and startPos are probably for 2 different points
    
    left_offset   = startDef
    left_slope    = 0
    kink_position = (finalPos+startPos)/2.0
    right_slope   = 2.0*(finalDef-startDef)/(finalPos-startPos)
    pstart = [left_offset, left_slope, kink_position, right_slope]

    offset_scale = (finalPos - startPos)/100
    left_slope_scale = right_slope/10
    kink_scale = (finalPos-startPos)/100
    right_slope_scale = right_slope
    scale = [offset_scale, left_slope_scale, kink_scale, right_slope_scale]
    if TEXT_VERBOSE or textVerbose :
        print "start Def %d, final def %d, start pos %d, final pos %d" % (startDef, finalDef, startPos, finalPos)
        print "Guessed params ", pstart
    def residual(p, y, x):
        Y = bilinear(x, p)
        return Y-y
    leastsq = scipy.optimize.leastsq
    params,cov,info,mesg,ier = leastsq(
        residual, pstart,
        args=(data["Deflection input"][dump_before_index:],
              data["Z piezo output"][dump_before_index:]),
        full_output=True, maxfev=10000)
    if TEXT_VERBOSE or textVerbose :
        print "Best fit parameters: ", params

    if PYLAB_VERBOSE or plotVerbose :
        _import_pylab()
        _pylab.figure(BASE_FIG_NUM+3)
        def plot_dict(d, label) :
            _pylab.plot(d["Z piezo output"], d["Deflection input"], '.',label=label)
        try :
            plot_dict(ddict['ret1'], 'First retract')
            plot_dict(ddict['mtpod'], 'Single step in')
            plot_dict(ddict['ret2'], 'Second retract')
        except KeyError :
            pass # if we don't have the surrounding data, don't worry
        plot_dict(data, 'Main approach')
        try :
            plot_dict(ddict['ret3'], 'Return to start')
        except KeyError :
            pass # if we don't have the surrounding data, don't worry
        def fit_fn(x, params) :
            if x <= params[2] :
                return params[0] + params[1]*x
            else :
                return params[0] + params[1]*params[2] + params[3]*(x-params[2])
        _pylab.plot([startPos, params[2], finalPos],
                    [fit_fn(startPos, params), fit_fn(params[2], params), fit_fn(finalPos, params)],
                    '-',label='fit')
        _pylab.title('surf_pos %5g %5g %5g %5g' % (params[0], params[1], params[2], params[3]))
        _pylab.legend(loc='best')
        _flush_plot()
    
    # check that the fit is reasonable
    # params[1] is slope in non-contact region
    # params[3] is slope in contact region
    if abs(params[1]*10) > abs(params[3]) :
        raise poorFit, "Slope in non-contact region, or no slope in contact"
    if params[2] < startPos+0.02*(finalPos-startPos) : # params[2] is kink position
        raise poorFit, "No kink (kink %g less than %g, need more space to left)" % \
            (params[2], startPos+0.02*(finalPos-startPos))
    if params[2] > finalPos-0.02*(finalPos-startPos) :
        raise poorFit, "No kink (kink %g more than %g, need more space to right)" % \
            (params[2], finalPos-0.02*(finalPos-startPos))
    if params[3] < 0.5 * right_slope : # params[3] is slope in contact region
        raise poorFit, "Too far"
    if TEXT_VERBOSE or textVerbose :
        print "\tSurface position: ", params[2]
    if retAllParams :
        return params
    return params[2]

def getSurfPos(zpiezo, maxDefl, textVerbose=False, plotVerbose=False) :
    ddict = getSurfPosData(zpiezo, maxDefl, textVerbose, plotVerbose)
    return analyzeSurfPosData(ddict, textVerbose, plotVerbose)

def test_smoothness(zp, plotVerbose=True) :
    posA = 20000
    posB = 50000
    setpoint = zp.def_V2in(3)
    steps = 200
    outfreq = 1e5
    outarray = linspace(posB, posA, 1000)
    indata=[]
    outdata=[]
    curVals = zp.jumpToPos(posA)
    zp.pCurVals(curVals)
    time.sleep(1) # let jitters die down
    for i in range(10) :
        print "ramp %d to %d" % (zp.curPos(), posB)
        curVals, data = moveToPosOrDef(zp, posB, setpoint, step=steps,
                                       return_data = True)
        indata.append(data)
        out = zp.ramp(outarray, outfreq)
        outdata.append(out)
    if plotVerbose :
        from pylab import figure, plot, title, legend, hold, subplot        
    if PYLAB_VERBOSE or plotVerbose :
        _import_pylab()
        _pylab.figure(BASE_FIG_NUM+4)
        for i in range(10) :
            _pylab.plot(indata[i]['Z piezo output'],
                        indata[i]['Deflection input'], '+--', label='in')
            _pylab.plot(outdata[i]['Z piezo output'],
                        outdata[i]['Deflection input'], '.-', label='out')
        _pylab.title('test smoothness (step in, ramp out)')
        #_pylab.legend(loc='best')
    
def test() :
    import z_piezo
    zp = z_piezo.z_piezo()
    curVals = zp.moveToPosOrDef(zp.pos_nm2out(600), defl=zp.curDef()+6000, step=(zp.pos_nm2out(10)-zp.pos_nm2out(0)))
    if TEXT_VERBOSE :
        zp.pCurVals(curVals)
    pos = zp.getSurfPos(maxDefl=zp.curDef()+6000)
    if TEXT_VERBOSE :
        print "Surface at %g nm", pos
    print "success"
    if PYLAB_VERBOSE and _final_flush_plot != None:
        _final_flush_plot()
